/* Copyright (C) 2006 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <string.h>
#include <stdio.h>
#include <memory.h>
#include "Engine.h"
#include "StorageHandler.h"
#include "StorageConnection.h"
#include "Connection.h"
#include "SyncObject.h"
#include "Sync.h"
#include "SQLError.h"
#include "StorageDatabase.h"
#include "StorageTableShare.h"
#include "Stream.h"
#include "Configuration.h"
#include "Threads.h"
#include "Connection.h"
#include "PreparedStatement.h"
#include "ResultSet.h"
#include "Database.h"
#include "InfoTable.h"


#define DICTIONARY_NAME			"falcon_dictionary"
#define DICTIONARY_PATH			"falcon_dictionary.fdd"
#define DICTIONARY_ACCOUNT		"mysql"
#define DICTIONARY_PW			"mysql"

#define HASH(address,size)				(((UIPTR) address >> 2) % size)

struct StorageSavepoint {
	StorageSavepoint*	next;
	StorageConnection*	storageConnection;
	int					savepoint;
	};

static const char *falconSchema [] = {
	"upgrade table falcon.tablespaces ("
	"    name varchar(128) not null primary	key,"
	"    pathname varchar(1024) not null)",
	
	"upgrade table falcon.tables ("
	"    given_schema_name varchar(128) not null,"
	"    effective_schema_name varchar(128) not null,"
	"    given_table_name varchar(128) not null,"
	"    effective_table_name varchar(128) not null,"
	"    tablespace_name varchar(128) not null,"
	"    pathname varchar(1024) not null primary key)",
	
	"upgrade unique index effective on falcon.tables (effective_schema_name, effective_table_name)",

	NULL };
				
class Server;
extern Server*	startServer(int port, const char *configFile);

static StorageHandler *storageHandler;

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#endif

StorageHandler*	getFalconStorageHandler(int lockSize)
{
	if (!storageHandler)
		storageHandler = new StorageHandler(lockSize);
	
	return storageHandler;
}


StorageHandler::StorageHandler(int lockSize)
{
	mySqlLockSize = lockSize;
	memset(connections, 0, sizeof(connections));
	memset(storageDatabases, 0, sizeof(storageDatabases));
	memset(tables, 0, sizeof(tables));
	dictionaryConnection = NULL;
	databaseList = NULL;
	globalTableSpace = false;
}

StorageHandler::~StorageHandler(void)
{
	for (int n = 0; n < databaseHashSize; ++n)
		for (StorageDatabase *storageDatabase; (storageDatabase = storageDatabases[n]);)
			{
			storageDatabases[n] = storageDatabase->collision;
			delete storageDatabase;
			}
	
	for (int n = 0; n < tableHashSize; ++n)
		for (StorageTableShare *table; (table = tables[n]);)
			{
			tables[n] = table->collision;
			delete table;
			}
}

void StorageHandler::startNfsServer(void)
{
	startServer(0, NULL);
}

void StorageHandler::addNfsLogger(int mask, Logger listener, void* arg)
{
	addLogListener(mask, listener, arg);
}

/***
StorageConnection* StorageHandler::getStorageConnection(const char* dbName, THD* mySqlThread, OpenOption createFlag, const char *path, const char *tableSpace)
{
	Sync sync(&syncObject, "StorageConnection::getStorageConnection");
	StorageDatabase *storageDatabase = NULL;
	
	if (tableSpace)
		{
		storageDatabase = findTablespace(tableSpace);
		
		if (!storageDatabase)
			return NULL;
		
		dbName = storageDatabase->filename;
		}
		
	sync.lock(Shared);
	int slot = HASH(mySqlThread, connectionHashSize);
	StorageConnection *storageConnection;
	
	for (storageConnection = connections[slot]; storageConnection; storageConnection = storageConnection->collision)
		if (storageConnection->mySqlThread == mySqlThread && (!dbName ||storageConnection->path == dbName))
			{
			storageConnection->addRef();
			
			return storageConnection;
			}

	sync.unlock();
	sync.lock(Exclusive);
	
	for (storageConnection = connections[slot]; storageConnection; storageConnection = storageConnection->collision)
		if (storageConnection->mySqlThread == mySqlThread && (!dbName || storageConnection->matches(dbName)))
			{
			storageConnection->addRef();
			
			return storageConnection;
			}
	
	if (!dbName)
		return NULL;

	storageConnection = new StorageConnection(this, dbName, mySqlThread);
	bool success = false;
	
	if (createFlag != CreateDatabase && createFlag != OpenTemporaryDatabase)
		try
			{
			storageConnection->connect();
			success = true;
			}
		catch (SQLException& exception)
			{
			//fprintf(stderr, "database open failed: %s\n", exception.getText());
			storageConnection->setErrorText(exception.getText());
			
			if (createFlag == OpenDatabase)
				{
				delete storageConnection;
				
				return NULL;
				}
			}
	
	if (!success && createFlag != OpenDatabase)
		try
			{
			storageConnection->create();
			}
		catch (SQLException& exception)
			{
			//fprintf(stderr, "database create failed: %s\n", exception.getText());
			delete storageConnection;
			
			return NULL;
			}
	
	storageConnection->collision = connections[slot];
	connections[slot] = storageConnection;
	
	return storageConnection;
}
***/

void StorageHandler::shutdownHandler(void)
{
	if (dictionaryConnection)
		{
		dictionaryConnection->commit();
		dictionaryConnection->shutdownDatabase();
		dictionaryConnection = NULL;
		}
		
	for (int n = 0; n < databaseHashSize; ++n)
		for (StorageDatabase *storageDatabase = storageDatabases[n]; storageDatabase; storageDatabase = storageDatabase->collision)
			storageDatabase->close();
}

void StorageHandler::databaseDropped(StorageDatabase *storageDatabase, StorageConnection* storageConnection)
{
	if (!storageDatabase && storageConnection)
		storageDatabase = storageConnection->storageDatabase;
		
	if (storageDatabase)
		{
		Sync syncHash(&hashSyncObject, "StorageHandler::dropDatabase");
		int slot = JString::hash(storageDatabase->name, databaseHashSize);
		syncHash.lock(Exclusive);
		StorageDatabase **ptr;
		
		for (ptr = storageDatabases + slot; *ptr; ptr = &(*ptr)->collision)
			if (*ptr == storageDatabase)
				{
				*ptr = storageDatabase->collision;
				break;
				}

		for (ptr = &databaseList; *ptr; ptr = &(*ptr)->next)
			if (*ptr == storageDatabase)
				{
				*ptr = storageDatabase->next;
				break;
				}

		syncHash.unlock();
		storageDatabase->release();
		}

	Sync sync(&syncObject, "StorageHandler::~dropDatabase");
	sync.lock(Exclusive);

	for (int n = 0; n < connectionHashSize; ++n)
		for (StorageConnection *cnct = connections[n]; cnct; cnct = cnct->collision)
			if (cnct != storageConnection)
				cnct->databaseDropped(storageDatabase);
			
	sync.unlock();
}

void StorageHandler::remove(StorageConnection* storageConnection)
{
	Sync sync(&syncObject, "StorageHandler::remove");
	sync.lock(Exclusive);
	removeConnection(storageConnection);			
}

int StorageHandler::commit(THD* mySqlThread)
{
	Sync sync(&syncObject, "StorageHandler::commit");
	sync.lock(Shared);
	int slot = HASH(mySqlThread, connectionHashSize);
	
	for (StorageConnection *connection = connections[slot]; connection; connection = connection->collision)
		if (connection->mySqlThread == mySqlThread)
			{
			int ret =connection->commit();
			
			if (ret)
				return ret;
			}
	
	return 0;
}

int StorageHandler::prepare(THD* mySqlThread, int xidSize, const UCHAR *xid)
{
	Sync sync(&syncObject, "StorageHandler::prepare");
	sync.lock(Shared);
	int slot = HASH(mySqlThread, connectionHashSize);
	
	for (StorageConnection *connection = connections[slot]; connection; connection = connection->collision)
		if (connection->mySqlThread == mySqlThread)
			{
			int ret = connection->prepare(xidSize, xid);
			
			if (ret)
				return ret;
			}
	
	return 0;
}

int StorageHandler::rollback(THD* mySqlThread)
{
	Sync sync(&syncObject, "StorageHandler::rollback");
	sync.lock(Shared);
	int slot = HASH(mySqlThread, connectionHashSize);
	
	for (StorageConnection *connection = connections[slot]; connection; connection = connection->collision)
		if (connection->mySqlThread == mySqlThread)
			{
			int ret =connection->rollback();
			
			if (ret)
				return ret;
			}
	
	return 0;
}

int StorageHandler::releaseVerb(THD* mySqlThread)
{
	Sync sync(&syncObject, "StorageHandler::releaseVerb");
	sync.lock(Shared);
	int slot = HASH(mySqlThread, connectionHashSize);
	
	for (StorageConnection *connection = connections[slot]; connection; connection = connection->collision)
		if (connection->mySqlThread == mySqlThread)
			connection->releaseVerb();
	
	return 0;
}

int StorageHandler::rollbackVerb(THD* mySqlThread)
{
	Sync sync(&syncObject, "StorageHandler::rollbackVerb");
	sync.lock(Shared);
	int slot = HASH(mySqlThread, connectionHashSize);
	
	for (StorageConnection *connection = connections[slot]; connection; connection = connection->collision)
		if (connection->mySqlThread == mySqlThread)
			connection->rollbackVerb();
	
	return 0;
}

int StorageHandler::savepointSet(THD* mySqlThread, void* savePoint)
{
	Sync sync(&syncObject, "StorageHandler::savepointSet");
	sync.lock(Shared);
	int slot = HASH(mySqlThread, connectionHashSize);
	StorageSavepoint *savepoints = NULL;
	
	for (StorageConnection *connection = connections[slot]; connection; connection = connection->collision)
		if (connection->mySqlThread == mySqlThread)
			{
			StorageSavepoint *savepoint = new StorageSavepoint;
			savepoint->next = savepoints;
			savepoints = savepoint;
			savepoint->storageConnection = connection;
			savepoint->savepoint = connection->savepointSet();
			}
	
	*((void**) savePoint) = savepoints;
	
	return 0;
}

int StorageHandler::savepointRelease(THD* mySqlThread, void* savePoint)
{
	Sync sync(&syncObject, "StorageHandler::savepointRelease");
	sync.lock(Shared);
	
	for (StorageSavepoint *savepoints = *(StorageSavepoint**) savePoint, *savepoint; 
		  (savepoint = savepoints);)
		{
		savepoint->storageConnection->savepointRelease(savepoint->savepoint);
		savepoints = savepoint->next;
		delete savepoint;
		}
		
	*((void**) savePoint) = NULL;

	return 0;
}

int StorageHandler::savepointRollback(THD* mySqlThread, void* savePoint)
{
	Sync sync(&syncObject, "StorageHandler::savepointRollback");
	sync.lock(Shared);
	
	for (StorageSavepoint *savepoints = *(StorageSavepoint**) savePoint, *savepoint; 
		   (savepoint = savepoints);)
		{
		savepoint->storageConnection->savepointRollback(savepoint->savepoint);
		savepoints = savepoint->next;
		delete savepoint;
		}
	
	*((void**) savePoint) = NULL;

	return 0;
}

StorageDatabase* StorageHandler::getStorageDatabase(const char* dbName, const char* path)
{
	Sync sync(&hashSyncObject, "StorageHandler::getStorageDatabase");
	int slot = JString::hash(dbName, databaseHashSize);
	StorageDatabase *storageDatabase;
	
	if (storageDatabases[slot])
		{
		sync.lock(Shared);
		
		if ( (storageDatabase = findDatabase(dbName)) )
			return storageDatabase;
			
		sync.unlock();
		}
		
	sync.lock(Exclusive);

	if ( (storageDatabase = findDatabase(dbName)) )
		return storageDatabase;
	
	storageDatabase = new StorageDatabase(this, dbName, path);
	storageDatabase->load();
	storageDatabase->collision = storageDatabases[slot];
	storageDatabases[slot] = storageDatabase;
	storageDatabase->addRef();
	storageDatabase->next = databaseList;
	databaseList = storageDatabase;
	
	return storageDatabase;
}

void StorageHandler::closeDatabase(const char* path)
{
	Sync sync(&hashSyncObject, "StorageHandler::getStorageDatabase");
	int slot = JString::hash(path, databaseHashSize);
	sync.lock(Exclusive);
	
	for (StorageDatabase *storageDatabase, **ptr = storageDatabases + slot; (storageDatabase = *ptr); ptr = &storageDatabase->collision)
		if (storageDatabase->filename == path)
			{
			*ptr = storageDatabase->collision;
			storageDatabase->close();
			storageDatabase->release();
			break;
			}
}

/***
const char* StorageHandler::getEngineStatus(THD* mySqlThread)
{
	StorageConnection *storageConnection = getStorageConnection(NULL, mySqlThread, OpenDatabase, NULL, NULL);
	
	if (!storageConnection || !storageConnection->connection)
		return NULL;
	
	storageConnection->lastErrorText = storageConnection->connection->analyze(analyzeTables | analyzeSpace);
	
	return storageConnection->lastErrorText;
}
***/

void StorageHandler::releaseText(const char* text)
{
	delete [] text;
}

int StorageHandler::commitByXID(int xidLength, const UCHAR* xid)
{
	return 0;
}

int StorageHandler::rollbackByXID(int xidLength, const UCHAR* xis)
{
	return 0;
}

Connection* StorageHandler::getDictionaryConnection(void)
{
	if (!dictionaryConnection)
		{
		Configuration *configuration = new Configuration(NULL);
		configuration->recordMemoryUpper = 100000;
		configuration->pageCacheSize = 1001024;
		dictionaryConnection = new Connection(configuration);
		Threads *threads = new Threads(NULL);
		
		try
			{
			dictionaryConnection->openDatabase(DICTIONARY_NAME, DICTIONARY_PATH, DICTIONARY_ACCOUNT, DICTIONARY_PW, NULL, threads);
			}
		catch (SQLException&)
			{
			try
				{
				dictionaryConnection->createDatabase(DICTIONARY_NAME, DICTIONARY_PATH, DICTIONARY_ACCOUNT, DICTIONARY_PW, threads);
				}
			catch (SQLException&)
				{
				dictionaryConnection->close();
				dictionaryConnection = NULL;
				throw;
				}
			}
		
		Statement *statement = dictionaryConnection->createStatement();
		
		for (const char **ddl = falconSchema; *ddl; ++ddl)
			statement->executeUpdate(*ddl);
		
		statement->close();
		dictionaryConnection->commit();
		}
	
	return dictionaryConnection;
}

int StorageHandler::createTablespace(const char* tableSpaceName, const char* filename)
{
	StorageDatabase *storageDatabase = NULL;
	JString tableSpace = JString::upcase(tableSpaceName);
	
	try
		{
		storageDatabase = getStorageDatabase(tableSpace, filename);
		storageDatabase->save();
		}
	catch (SQLException&)
		{
		if (storageDatabase)
			{
			databaseDropped(storageDatabase, NULL);
			storageDatabase->dropDatabase();
			}
		}
	
	return 0;
}

int StorageHandler::deleteTablespace(const char* tableSpaceName)
{
	return 0;
}

StorageTableShare* StorageHandler::findTable(const char* pathname)
{
	Sync sync(&hashSyncObject, "StorageHandler::findTable");
	int slot = JString::hash(pathname, tableHashSize);
	StorageTableShare *tableShare;

	if (tables[slot])
		{
		sync.lock(Shared);
		
		for (tableShare = tables[slot]; tableShare; tableShare = tableShare->collision)
			if (tableShare->pathName == pathname)
				return tableShare;
	
		sync.unlock();
		}

	sync.lock(Exclusive);
	
	for (tableShare = tables[slot]; tableShare; tableShare = tableShare->collision)
		if (tableShare->pathName == pathname)
			return tableShare;
	
	tableShare = new StorageTableShare(this, pathname, NULL, mySqlLockSize, false);
	tableShare->collision = tables[slot];
	tables[slot] = tableShare;
	
	return tableShare;	
}

StorageTableShare* StorageHandler::createTable(const char* pathname, const char *tableSpaceName, bool tempTable)
{
	StorageTableShare *tableShare = new StorageTableShare(this, pathname, tableSpaceName, mySqlLockSize, tempTable);
	addTable(tableShare);
	tableShare->registerTable();
	
	return tableShare;	
}

void StorageHandler::addTable(StorageTableShare* table)
{
	int slot = JString::hash(table->pathName, tableHashSize);
	Sync sync(&hashSyncObject, "StorageHandler::add");
	sync.lock(Exclusive);
	table->collision = tables[slot];
	tables[slot] = table;
}

void StorageHandler::removeTable(StorageTableShare* table)
{
	Sync sync(&hashSyncObject, "StorageHandler::deleteTable");
	sync.lock(Exclusive);
	int slot = JString::hash(table->pathName, tableHashSize);
	
	for (StorageTableShare **ptr = tables + slot; *ptr; ptr = &(*ptr)->collision)
		if (*ptr == table)
			{
			*ptr = table->collision;
			break;
			}
}

StorageConnection* StorageHandler::getStorageConnection(StorageTableShare* tableShare, THD* mySqlThread, int mySqlThdId, OpenOption createFlag)
{
	Sync sync(&syncObject, "StorageConnection::getStorageConnection");
	
	if (!tableShare->storageDatabase)
		tableShare->findDatabase();

	int slot = HASH(mySqlThread, connectionHashSize);
	StorageConnection *storageConnection;
	
	if (connections[slot])
		{
		for (storageConnection = connections[slot]; storageConnection; storageConnection = storageConnection->collision)
			if (storageConnection->mySqlThread == mySqlThread && storageConnection->storageDatabase == tableShare->storageDatabase)
				{
				storageConnection->addRef();
				
				return storageConnection;
				}

		sync.lock(Shared);
		sync.unlock();
		}
		
	sync.lock(Exclusive);
	
	for (storageConnection = connections[slot]; storageConnection; storageConnection = storageConnection->collision)
		if (storageConnection->mySqlThread == mySqlThread && storageConnection->storageDatabase == tableShare->storageDatabase)
			{
			storageConnection->addRef();
			
			return storageConnection;
			}
	
	const char *dbName = tableShare->tableSpace;
	char path[FILENAME_MAX];
	tableShare->getDefaultPath(path);
	
	StorageDatabase *storageDatabase = getStorageDatabase(dbName, path);
	storageConnection = new StorageConnection(this, storageDatabase, mySqlThread, mySqlThdId);
	tableShare->setDatabase(storageDatabase);
	storageDatabase->release();
	bool success = false;
	
	if (createFlag != CreateDatabase && createFlag != OpenTemporaryDatabase)
		try
			{
			storageConnection->connect();
			success = true;
			}
		catch (SQLException& exception)
			{
			//fprintf(stderr, "database open failed: %s\n", exception.getText());
			storageConnection->setErrorText(exception.getText());
			
			if (createFlag == OpenDatabase)
				{
				delete storageConnection;
				
				return NULL;
				}
			}
	
	if (!success && createFlag != OpenDatabase)
		try
			{
			storageConnection->create();
			}
		catch (SQLException&)
			{
			delete storageConnection;
			
			return NULL;
			}
	
	storageConnection->collision = connections[slot];
	connections[slot] = storageConnection;
	
	return storageConnection;
}

StorageDatabase* StorageHandler::findDatabase(const char* dbName)
{
	int slot = JString::hash(dbName, databaseHashSize);
	
	for (StorageDatabase *storageDatabase = storageDatabases[slot]; storageDatabase; storageDatabase = storageDatabase->collision)
		if (storageDatabase->name == dbName)
			{
			storageDatabase->addRef();
			
			return storageDatabase;
			}
			
	return NULL;
}

void StorageHandler::changeMySqlThread(StorageConnection* storageConnection, THD* newThread)
{
	Sync sync(&syncObject, "StorageHandler::changeMySqlThread");
	sync.lock(Exclusive);
	removeConnection(storageConnection);
	storageConnection->mySqlThread = newThread;		
	int slot = HASH(storageConnection->mySqlThread, connectionHashSize);
	storageConnection->collision = connections[slot];
	connections[slot] = storageConnection;	
}

void StorageHandler::removeConnection(StorageConnection* storageConnection)
{
	int slot = HASH(storageConnection->mySqlThread, connectionHashSize);
	
	for (StorageConnection **ptr = connections + slot; *ptr; ptr = &(*ptr)->collision)
		if (*ptr == storageConnection)
			{
			*ptr = storageConnection->collision;
			break;
			}
}

int StorageHandler::closeConnections(THD* thd)
{
	int slot = HASH(thd, connectionHashSize);
	Sync sync(&syncObject, "StorageHandler::closeConnections");
	
	for (StorageConnection *storageConnection = connections[slot], *next; storageConnection; storageConnection = next)
		{
		next = storageConnection->collision;
		
		if (storageConnection->mySqlThread == thd)
			{
			storageConnection->close();
		
			if (storageConnection->mySqlThread)
				storageConnection->release();	// This is for thd->ha_data[falcon_hton->slot]
			
			storageConnection->release();	// This is for storageConn
			}
		}
	
	return 0;
}

int StorageHandler::dropDatabase(const char* path)
{
	char pathname[FILENAME_MAX];
	const char *slash = pathname;
	char *q = pathname;
	
	for (const char *p = path; *p;)
		{
		char c = *p++;
		
		if (c == '/')
			{
			if (*p == 0)
				break;
				
			slash = q + 1;
			}
		
		*q++ = c;
		}
	
	*q = 0;
	JString dbName = JString::upcase(slash);
	strcpy(q, StorageTableShare::getDefaultRoot());	
	StorageDatabase *storageDatabase = getStorageDatabase(dbName, pathname);
	databaseDropped(storageDatabase, NULL);

	try
		{
		storageDatabase->dropDatabase();
		}
	catch (SQLException&)
		{
		}

	storageDatabase->release();
	
	return 0;
}

void StorageHandler::getIOInfo(InfoTable* infoTable)
{
	Sync sync(&hashSyncObject, "StorageHandler::getIOInfo");
	sync.lock(Shared);
	
	for (StorageDatabase *storageDatabase = databaseList; storageDatabase; storageDatabase = storageDatabase->next)
		storageDatabase->masterConnection->database->getIOInfo(infoTable);
}

void StorageHandler::getMemoryDetailInfo(InfoTable* infoTable)
{
	MemMgrAnalyze(MemMgrSystemDetail, infoTable);
}

void StorageHandler::getMemorySummaryInfo(InfoTable* infoTable)
{
	MemMgrAnalyze(MemMgrSystemSummary, infoTable);
}

void StorageHandler::getRecordCacheDetailInfo(InfoTable* infoTable)
{
	MemMgrAnalyze(MemMgrRecordDetail, infoTable);
}

void StorageHandler::getRecordCacheSummaryInfo(InfoTable* infoTable)
{
	MemMgrAnalyze(MemMgrRecordSummary, infoTable);
}

void StorageHandler::getTransactionInfo(InfoTable* infoTable)
{
	Sync sync(&hashSyncObject, "StorageHandler::getTransactionInfo");
	sync.lock(Shared);
	
	for (StorageDatabase *storageDatabase = databaseList; storageDatabase; storageDatabase = storageDatabase->next)
		if (storageDatabase->masterConnection && storageDatabase->masterConnection->database)
			storageDatabase->masterConnection->database->getTransactionInfo(infoTable);
}

void StorageHandler::getSerialLogInfo(InfoTable* infoTable)
{
	Sync sync(&hashSyncObject, "StorageHandler::getTransactionInfo");
	sync.lock(Shared);
	
	for (StorageDatabase *storageDatabase = databaseList; storageDatabase; storageDatabase = storageDatabase->next)
		if (storageDatabase->masterConnection && storageDatabase->masterConnection->database)
			storageDatabase->masterConnection->database->getSerialLogInfo(infoTable);
}

void StorageHandler::getSyncInfo(InfoTable* infoTable)
{
	SyncObject::getSyncInfo(infoTable);
}

void StorageHandler::getTransactionSummaryInfo(InfoTable* infoTable)
{
	Sync sync(&hashSyncObject, "StorageHandler::getTransactionSummaryInfo");
	sync.lock(Shared);
	
	for (StorageDatabase *storageDatabase = databaseList; storageDatabase; storageDatabase = storageDatabase->next)
		if (storageDatabase->masterConnection && storageDatabase->masterConnection->database)
			storageDatabase->masterConnection->database->getTransactionSummaryInfo(infoTable);
}

